@@ -7,18 +7,25 @@
 #include "constants.h"
 #include "datrie.h"
 #include "utils.h"
+#include "zstdfilter.h"
 #include <boost/algorithm/cxx11/all_of.hpp>
 #include <boost/algorithm/string.hpp>
+#include <boost/iostreams/filtering_streambuf.hpp>
 #include <boost/range/adaptor/reversed.hpp>
 #include <boost/range/adaptor/transformed.hpp>
 #include <boost/range/algorithm.hpp>
 #include <cmath>
 #include <fcitx-utils/log.h>
+#include <fcitx-utils/stringutils.h>
+#include <iterator>
+#include <ostream>
+#include <stdexcept>
+#include <string>
 
 namespace libime {
 
 static constexpr uint32_t historyBinaryFormatMagic = 0x000fc315;
-static constexpr uint32_t historyBinaryFormatVersion = 0x2;
+static constexpr uint32_t historyBinaryFormatVersion = 0x3;
 
 struct WeightedTrie {
     using TrieType = DATrie<int32_t>;
@@ -157,6 +164,22 @@ public:
         }
     }
 
+    void loadText(std::istream &in) {
+        clear();
+        std::string buf;
+        std::vector<std::string> lines;
+        while (std::getline(in, buf)) {
+            lines.emplace_back(buf);
+            if (lines.size() >= maxSize_)
+                break;
+        }
+        for (auto &line : lines | boost::adaptors::reversed) {
+            std::vector<std::string> sentence =
+                fcitx::stringutils::split(line, " ");
+            add(sentence);
+        }
+    }
+
     void save(std::ostream &out) {
         uint32_t count = recent_.size();
         throw_if_io_fail(marshall(out, count));
@@ -200,6 +223,13 @@ public:
         if (sentence.empty()) {
             return popedSentence;
         }
+        // Validate data.
+        if (std::any_of(std::begin(sentence), std::end(sentence),
+                        [](const std::string &word) {
+                            return word.find('\0') != std::string::npos;
+                        })) {
+            return popedSentence;
+        }
         while (recent_.size() >= maxSize_) {
             remove(recent_.back());
             popedSentence.splice(popedSentence.end(), recent_,
@@ -373,6 +403,8 @@ HistoryBigram::HistoryBigram()
     FCITX_D();
     const float p = 1.0 / (1 + HISTORY_BIGRAM_ALPHA_VALUE);
     constexpr std::array<int, 3> poolSize = {128, 8192, 65536};
+    d->pools_.reserve(poolSize.size());
+    d->poolWeight_.reserve(poolSize.size());
     for (auto size : poolSize) {
         d->pools_.emplace_back(size);
         float portion = 1.0f;
@@ -441,7 +473,7 @@ float HistoryBigram::score(std::string_view prev, std::string_view cur) const {
     auto bf = d->bigramFreq(prev, cur);
     auto uf1 = d->unigramFreq(cur);
 
-    float bigramWeight = d->useOnlyUnigram_ ? 0.0f : 0.68f;
+    float bigramWeight = d->useOnlyUnigram_ ? 0.0f : 0.8f;
     // add 0.5 to avoid div 0
     float pr = 0.0f;
     pr += bigramWeight * float(bf) / float(uf0 + d->poolWeight_[0] / 2);
@@ -472,20 +504,47 @@ void HistoryBigram::load(std::istream &in) {
         std::for_each(d->pools_.begin(), d->pools_.begin() + 2,
                       [&in](auto &pool) { pool.load(in); });
         break;
-    case historyBinaryFormatVersion:
+    case 2:
         boost::range::for_each(d->pools_, [&in](auto &pool) { pool.load(in); });
         break;
+    case historyBinaryFormatVersion: {
+        boost::iostreams::filtering_istreambuf compressBuf;
+        compressBuf.push(ZSTDDecompressor());
+        compressBuf.push(in);
+        std::istream compressIn(&compressBuf);
+
+        boost::range::for_each(
+            d->pools_, [&compressIn](auto &pool) { pool.load(compressIn); });
+        // We don't want to read any data, but only trigger the zstd footer
+        // handling, which validates CRC.
+        compressIn.peek();
+        if (compressIn.bad()) {
+            throw std::invalid_argument("Failed to validate the history data");
+        }
+        break;
+    }
     default: {
         throw std::invalid_argument("Invalid history version.");
     }
     }
 }
 
+void HistoryBigram::loadText(std::istream &in) {
+    FCITX_D();
+    boost::range::for_each(d->pools_, [&in](auto &pool) { pool.loadText(in); });
+}
+
 void HistoryBigram::save(std::ostream &out) {
     FCITX_D();
     throw_if_io_fail(marshall(out, historyBinaryFormatMagic));
     throw_if_io_fail(marshall(out, historyBinaryFormatVersion));
-    boost::range::for_each(d->pools_, [&out](auto &pool) { pool.save(out); });
+    boost::iostreams::filtering_streambuf<boost::iostreams::output> compressBuf;
+    compressBuf.push(ZSTDCompressor());
+    compressBuf.push(out);
+    std::ostream compressOut(&compressBuf);
+
+    boost::range::for_each(
+        d->pools_, [&compressOut](auto &pool) { pool.save(compressOut); });
 }
 
 void HistoryBigram::dump(std::ostream &out) {
